/************************************************
 **  EMUTAG SDK EXAMPLE FOR NTAG203 EMULATION  **
 ************************************************/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include "main.h"

#define C_REQA		0x26
#define C_WUPA		0x52
#define C_SEL1		0x93
#define C_SEL2		0x95
#define C_HALT		0x50
#define C_READ		0x30
#define C_WRITE		0xA2
#define C_COMPAT_WRITE	0xA0

#define R_ATQA_H	0x00
#define R_ATQA_L	0x44
#define R_SAK		0x00
#define R_CT		0x88

// 4-bit responses are left-shifted by 4 bits, as data is shifted out left MSB first
#define R_ACK		0xA0
#define R_ERR		0x10
#define R_NAK		0x00

#define S_IDLE		0
#define S_READY1	1
#define S_READY2	2
#define S_ACTIVE	3
#define S_COMPAT_WRITE	4

#define BL_OTP		0
#define BL_94		1
#define BL_FA		2
#define BL_13		0
#define BL_57		4
#define BL_CNT		0

#define NUM_PAGES	42

// Simulation of write delay to internal EEPROM: 4.1 ms fixed from request
// TIMER0 is counting with prescaler = 1024 with count reset after receiving incoming message
// bitrate = F_CPU / 128, TIMER0 prescaler = 1024, ratio = 8
// 4.1 ms / (9 bit char / bitrate) = 48 byte frames
// (48 frames * 9 - 3 (turnaround bit buffer compensation)) / 8 = 53 timer counts
#define WRITE_DELAY	53

// registers contents are saved between function calls, except for shared registers
register volatile uint8_t lock_sw  asm("r7");
register volatile uint8_t lock_b0  asm("r8");
register volatile uint8_t lock_b1  asm("r9");
register volatile uint8_t lock_b2  asm("r10");
register volatile uint8_t lock_b3  asm("r11");

uint8_t mem_array[NUM_PAGES * 4]; // namely the NTAG203 memory storage array

uint8_t perm_tab[8] PROGMEM = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80}; // used by write_perm()

/* MACROS */

#define SEND_ERR(code)				\
	state = S_IDLE;				\
	tx_buf[0] = code;			\
	tx_end = 1;				\
	tx_bits = 4;				\
	ctrl_flags |= 1 << F_TX_PAR_OFF;	\
	ctrl_flags |= 1 << F_TX_ENABLE;		\
	goto user_end;

/* FUNCTIONS */

/*  user_init() : Called from main code once at boot time. Can be used to initialize memory.
 *              : This function is required to be present. Leave empty if not used.
 */
void user_init(void);

/*  user_proc() : Called from main code when an incoming message has been received and stored in tx_buf[].
 *
 *  rx_bytes      = number of complete bytes stored in tx_buf[]
 *  rx_bits       = number of additional bits received if the last byte is incomplete
 *  rx_bits_total = rx_bytes * 9 + rx_bits, cached for faster comparisons
 *  __attribute__((OS_main)) tells the compiler not to save any registers on stack, as main code reloads everything
 */
void user_proc(uint8_t rx_bytes, uint8_t rx_bits, uint8_t rx_bits_total) __attribute__((OS_main));

/*  write_perm() : Computes whether ACK or NAK should be returned after WRITE or COMPATIBILITY_WRITE commands.
 *               : The result depends on page, lock-bits, block-locking bits, and lock switch position.
 *
 *  page = number of 4-byte block in memory for which write permissions are to be checked
 *  returns 0 if NAK needs to be returned, 1 if ACK
 */
uint8_t write_perm(uint8_t page);

/*  counter() : Updates software one-way counter value in mem_array[] using data from rx_buf[].
 *
 *  src = offset in rx_buf[] to start transferring data from
 *  returns 0 if counter is 0xFFFF or overflows (NAK), otherwise 1 (ACK) and the counter value is updated with delay.
 */
uint8_t counter(uint8_t src);

/*  buf_save() : Writes bytes from rx_buf[] to mem_array[] in WRITE or COMPATIBILITY_WRITE commands.
 *             : Data is transferred according to rules defined in MIFARE Ultralight write access conditions.
 *             : E.g. page 3 (OTP) is OR'ed with incoming data in locked mode, and overwritten in unlocked mode.
 *
 *  page = number of 4-byte block in memory which needs to be written with incoming data
 *  src  = offset in rx_buf[] to start transferring data from
 */
void buf_save(uint8_t page, uint8_t src);

/*  buf_copy() : Copies bytes from mem_array[] to tx_buf[]. Used in READ command and anticolision procedure.
 *
 *  dst = offset in tx_buf[] to start copying to
 *  src = offset in mem_array[] to start copying from
 *  len = number of bytes to copy
 */
void buf_copy(uint8_t dst, uint8_t src, uint8_t len) __attribute__((noinline));

/*  buf_cmp() : Compares bytes and bits in rx_buf[] and tx_buf[] during anticollision.
 *  Two first bytes in both buffers are skipped (not compared), since they contain opcode and NVB.
 *
 *  len  = total number of complete bytes in rx_buf[]
 *  bits = number of additional bits in rx_buf[]
 *  returns 0 if all bits of partial UID match in rx_buf[] and tx_buf[], 1 otherwise
 */
uint8_t buf_cmp(uint8_t len, uint8_t bits);

void user_init(void) {
	mem_array[0x00] = 0x04;
	mem_array[0x03] = 0x8c;
	mem_array[0x09] = 0x48;
	mem_array[0x0c] = 0xe1;
	mem_array[0x0d] = 0x10;
	mem_array[0x0e] = 0x12;
	
	mem_array[0x10] = 0x01;
	mem_array[0x11] = 0x03;
	mem_array[0x12] = 0xa0;
	mem_array[0x13] = 0x10;
	mem_array[0x14] = 0x44;
	mem_array[0x15] = 0x03;
	mem_array[0x17] = 0xfe;
	
	lock_b0 = lock_b1 = lock_b2 = lock_b3 = 0;
}

void user_proc(uint8_t rx_bytes, uint8_t rx_bits, uint8_t rx_bits_total) {
	static uint8_t compat_wr_addr;

	uint8_t op, arg, old_state;
	uint8_t tx_start = 0, tx_bits = 0, tx_end = 0; // initialize arguments for reply()
	
	if(ctrl_flags & 1 << F_RX_CHR_ERR) goto user_end;
	if(ctrl_flags & 1 << F_RX_OVF_ERR) goto user_end;
	
	// COMPATIBILITY_WRITE command consists of 2 request/response frames, need to save as special state
	if(state == S_COMPAT_WRITE) {
		if(rx_bits) goto user_end;
		if(rx_bytes != 18) goto user_end;
		
		if((ctrl_flags & 1 << F_RX_PAR_ERR) || crc_chk(rx_bytes)) {
			state = S_IDLE;
			tx_buf[0] = R_ERR;
		}
		else if(compat_wr_addr == 41 && (lock_sw & 1 << LOCK_PIN)) {
			if(!counter(0)) {
				state = S_IDLE;
				tx_buf[0] = R_NAK;
			}
			else {
				state = S_ACTIVE;
				tx_buf[0] = R_ACK;
			}
		}
		else if(compat_wr_addr == 1 && rx_buf[0] == R_CT) {
			// prevent storing cascade tag in UID3
			// Level 3 (10-byte) UIDs not supported
			state = S_IDLE;
			tx_buf[0] = R_NAK;
		}
		else {
			buf_save(compat_wr_addr, 0);
			state = S_ACTIVE;
			tx_buf[0] = R_ACK;
		}
		
		tx_end = 1;
		tx_bits = 4;
		ctrl_flags |= 1 << F_TX_PAR_OFF;
		ctrl_flags |= 1 << F_TX_ENABLE;
	}
	// other commands or first part of COMPATIBILITY_WRITE command
	else {
		op = rx_buf[0];
		
		// REQA / WUPA frames
		if(rx_bits_total == 7) {
			if(op == C_REQA || op == C_WUPA) {
				if(state == S_IDLE && (op == C_WUPA || (~ctrl_flags & 1 << F_ST_HALT))) {
					// activate new lock config
					lock_b0 = mem_array[0x0a];
					lock_b1 = mem_array[0x0b];
					lock_b2 = mem_array[0xa0];
					lock_b3 = mem_array[0xa1];
					
					state = S_READY1;
					tx_buf[0] = R_ATQA_L;
					tx_buf[1] = R_ATQA_H;
					tx_end = 2;
					ctrl_flags |= 1 << F_TX_ENABLE;
				}
				else state = S_IDLE;
			}
			else goto user_end;
		}
		
		// other frames
		if(rx_bits_total >= 18) {
			if(ctrl_flags & 1 << F_RX_PAR_ERR) { SEND_ERR(R_ERR); }
			
			arg = rx_buf[1];
			
			// ANTICOLLISION / SELECT commands using anticollision frames
			if(op == C_SEL1 || op == C_SEL2) {
				if(arg == 0x70) {
					if(crc_chk(rx_bytes)) { SEND_ERR(R_ERR); }
					rx_bytes -= 2;
				}
				
				// Check NVB for valid range
				if(arg & 8) goto user_end;
				if(arg < 0x20 || arg > 0x70) goto user_end;
				
				// Check NVB against actual data length
				if((arg >> 4) != rx_bytes) goto user_end;
				if((arg & 0x0f) != rx_bits) goto user_end;
				
				old_state = state;
				state = S_IDLE;
				
				if(op == C_SEL1) {
					tx_buf[2] = R_CT;
					buf_copy(3, 0, 4); // extract CT UID0 UID1 UID2 BCC0
				}
				else buf_copy(2, 4, 5); // extract UID3 UID4 UID5 UID6 BCC1
				
				if(buf_cmp(rx_bytes, rx_bits)) goto user_end;
				
				if(old_state == S_READY2 || (old_state == S_READY1 && op == C_SEL1)) {
					state = old_state;
					if(arg == 0x70) {
						state++;
						if(op == C_SEL1) tx_buf[0] = R_SAK | 0x04; // cascade bit
						else             tx_buf[0] = R_SAK;
						tx_end = 1;
						ctrl_flags |= 1 << F_TX_CRC;
					}
					else {
						tx_start = rx_bytes;
						tx_bits = rx_bits;
						tx_end = 7;
					}
					ctrl_flags |= 1 << F_TX_ENABLE;
				}
			}
			
			// other commands using standard frames
			else {
				if(rx_bits) goto user_end;
				if(crc_chk(rx_bytes)) { SEND_ERR(R_ERR); }
				rx_bytes -= 2;
				
				if(op == C_HALT) {
					if(rx_bytes != 2) goto user_end;
					if(arg) goto user_end;
					if(state == S_ACTIVE) ctrl_flags |= 1 << F_ST_HALT;
					state = S_IDLE;
				}
				
				if(op == C_READ) {
					if(rx_bytes != 2) goto user_end;
					old_state = state;
					state = S_IDLE;
					
					if((!arg && old_state != S_IDLE) || (arg < NUM_PAGES && old_state == S_ACTIVE)) {
						state = S_ACTIVE;
						
						arg <<= 2;
						for(tx_end = 0; tx_end < 16; tx_end++) {
							tx_buf[tx_end] = mem_array[arg];
							arg++;
							if(arg == NUM_PAGES * 4) arg = 0; // wrap over memory array
						}
						
						ctrl_flags |= 1 << F_TX_CRC;
						ctrl_flags |= 1 << F_TX_ENABLE;
					}
					else if(arg >= NUM_PAGES && old_state == S_ACTIVE) {
						//state = S_ACTIVE; // RESET to IDLE after failed read!
						tx_buf[0] = R_NAK;
						tx_end = 1;
						rx_bits = 4;
						ctrl_flags |= 1 << F_TX_PAR_OFF;
						ctrl_flags |= 1 << F_TX_ENABLE;
					}
				}
				
				if(op == C_WRITE || op == C_COMPAT_WRITE) {
					if(op == C_WRITE) tx_end = 6;
					else              tx_end = 2;
					
					if(rx_bytes != tx_end) goto user_end;
					if(state != S_ACTIVE) {
						state = S_IDLE;
						goto user_end;
					}
					
					// strobe lock switch position into register
					asm volatile("in %0, %1" : "=r" (lock_sw) : "I" (_SFR_IO_ADDR(LOCK_PORT)));
					
					if(arg < NUM_PAGES && write_perm(arg)) {
						tx_buf[0] = R_ACK;
						
						if(op == C_WRITE) {
							if(arg == 41 && (lock_sw & 1 << LOCK_PIN)) {
								if(!counter(2)) {
									tx_buf[0] = R_NAK;
									state = S_IDLE;
								}
							}
							else if(arg == 1 && rx_buf[2] == R_CT) {
								// prevent storing cascade tag in UID3
								// Level 3 (10-byte) UIDs not supported
								tx_buf[0] = R_NAK;
								state = S_IDLE; // RESET to IDLE after failed write!
							}
							else buf_save(arg, 2);
						}
						else {
							state = S_COMPAT_WRITE;
							compat_wr_addr = arg;
						}
					}
					else {
						tx_buf[0] = R_NAK;
						state = S_IDLE; // RESET to IDLE after failed write!
					}
					
					tx_end = 1;
					tx_bits = 4;
					ctrl_flags |= 1 << F_TX_PAR_OFF;
					ctrl_flags |= 1 << F_TX_ENABLE;
				}
			}
		}
	}
	
	user_end:
	
	reply(tx_start, tx_bits, tx_end);
	// reply() must be called at the end of user_proc(), even if nothing is to be transmitted!
}	

uint8_t write_perm(uint8_t page) {
	if(~lock_sw & 1 << LOCK_PIN) return 1;
	
	if(page < 16) {
		if(page < 3) {
			if(page < 2)  { return 0; }
			else          { return 1; }
		}
		else {
			if(page < 8)  { if(lock_b0 & pgm_read_byte(&(perm_tab[page])))     return 0; }
			else          { if(lock_b1 & pgm_read_byte(&(perm_tab[page & 7]))) return 0; }
		}
	}
	else {
		if(page < 40) {
			if(page < 28) { if(lock_b2 & pgm_read_byte(&(perm_tab[(page >> 2) - 3]))) return 0; }
			else          { if(lock_b2 & pgm_read_byte(&(perm_tab[(page >> 2) - 2]))) return 0; }
		}
		else {
			if(page < 41) { return 1; }
			else          { if(lock_b3 & 0b00010000) return 0; }
		}
	}
	
	return 1;
}

uint8_t counter(uint8_t src) {
	union { uint16_t u16; uint8_t u8[2]; } cnt;
	uint8_t *src_ptr;
	src_ptr = rx_buf + src;
	
	cnt.u8[0] = mem_array[0xa4];
	cnt.u8[1] = mem_array[0xa5];
	
	if(cnt.u16 == 0xffff) return 0;
	
	if(cnt.u16) {
		cnt.u16 += *src_ptr & 0x0f;
		if(cnt.u8[1] < mem_array[0xa5]) return 0; // overflow
		
		mem_array[0xa4] = cnt.u8[0];
		mem_array[0xa5] = cnt.u8[1];
	}
	else {
		mem_array[0xa4] = *(src_ptr+0);
		mem_array[0xa5] = *(src_ptr+1);
	}
	
	while(TCNT0 < WRITE_DELAY);
	return 1;
}

void buf_save(uint8_t page, uint8_t src) {
	uint8_t *dst_ptr, *src_ptr, tmp = 4;
	dst_ptr = mem_array + (page << 2);
	src_ptr = rx_buf + src;
	
	if(page == 2) { *(src_ptr+0) = *(dst_ptr+0); } // do not update BCC1
	
	if(lock_sw & 1 << LOCK_PIN) {
		if(page == 2)  { *(src_ptr+1) = *(dst_ptr+1); } // do not update internal byte
		if(page == 40) {
			*(src_ptr+2) = *(dst_ptr+2);
			*(src_ptr+3) = *(dst_ptr+3); // do not update unused bytes
		}
		
		if(page == 40 || (page & ~1) == 2) {
			while(tmp) {
				*src_ptr++ |= *dst_ptr++;
				asm volatile("dec %0" : "=r" (tmp) : "0" (tmp));
			}
			dst_ptr -= 4;
			src_ptr -= 4;
			tmp = 4;
		}
	}
	
	// assembler is used below to speed-up pointer manipulation and comparison while copying
	// see assembler listing file to check compilation result if not sure
	
	while(tmp) {
		*dst_ptr++ = *src_ptr++;
		asm volatile("dec %0" : "=r" (tmp) : "0" (tmp));
	}
	dst_ptr -= 4;
	src_ptr -= 4;
	
	// revert bits locked by block-locking bits
	if(lock_sw & 1 << LOCK_PIN) {
		if(page == 2) {
			if(lock_b0 & 1 << BL_OTP) {
				tmp = *(dst_ptr+2);
				// bit transfer instructions below: copy bit 3 from lock_b0 to tmp
				asm volatile(
					"bst	%2, 3		\n\t"
					"bld	%0, 3		\n\t"
					: "=r" (tmp) : "0" (tmp), "r" (lock_b0)
				);
				*(dst_ptr+2) = tmp;
			}
			
			if(lock_b0 & 1 << BL_94) {
				*(dst_ptr+2) |= lock_b0 & 0b11110000;
				*(dst_ptr+2) &= lock_b0 | 0b00001111;
				*(dst_ptr+3) |= lock_b1 & 0b00000011;
				*(dst_ptr+3) &= lock_b1 | 0b11111100;
			}
			
			if(lock_b0 & 1 << BL_FA) {
				*(dst_ptr+3) |= lock_b1 & 0b11111100;
				*(dst_ptr+3) &= lock_b1 | 0b00000011;
			}
		}
		
		if(page == 40) {
			if(lock_b3 & 1 << BL_CNT) {
				tmp = *(dst_ptr+1);
				// bit transfer instructions below: copy bit 4 from lock_b3 to tmp
				asm volatile(
					"bst	%2, 4		\n\t"
					"bld	%0, 4		\n\t"
					: "=r" (tmp) : "0" (tmp), "r" (lock_b3)
				);
				*(dst_ptr+1) = tmp;
			}
			
			if(lock_b2 & 1 << BL_13) {
				*(dst_ptr+0) |= lock_b2 & 0b00001110;
				*(dst_ptr+0) &= lock_b2 | 0b11110001;
			}
			
			if(lock_b2 & 1 << BL_57) {
				*(dst_ptr+0) |= lock_b2 & 0b11100000;
				*(dst_ptr+0) &= lock_b2 | 0b00011111;
			}
		}
	}
	
	// recompute BCC0 and BCC1
	if(page == 1) *(dst_ptr+4) = *(src_ptr+0) ^ *(src_ptr+1) ^ *(src_ptr+2) ^ *(src_ptr+3);
	if(page == 0) *(dst_ptr+3) = *(src_ptr+0) ^ *(src_ptr+1) ^ *(src_ptr+2) ^ R_CT;
	
	while(TCNT0 < WRITE_DELAY);
}

void buf_copy(uint8_t dst, uint8_t src, uint8_t len) {
	uint8_t *dst_ptr, *src_ptr;
	dst_ptr = tx_buf + dst;
	src_ptr = mem_array + src;
	while(len) {
		*dst_ptr++ = *src_ptr++;
		asm volatile("dec %0" : "=r" (len) : "0" (len)); // speed up copying using assembler...
	}
}

uint8_t buf_cmp(uint8_t len, uint8_t bits) {
	uint8_t mask = 0xff, tmp1, tmp2;
	uint8_t *ptr1, *ptr2;
	ptr1 = rx_buf + 2;
	ptr2 = tx_buf + 2;
	len -= 2;
	
	// speed up comparison using assembler...
	while(len) {
		asm volatile("ld %0, X+" : "=r" (tmp1), "=x" (ptr1) : "1" (ptr1));
		asm volatile("ld %0, Z+" : "=r" (tmp2), "=z" (ptr2) : "1" (ptr2));
		if(tmp1 != tmp2) return 1;
		asm volatile("dec %0" : "=r" (len) : "0" (len));
	}
	
	// replace default implementation of shifting by multiple bits:
	// mask <<= bits;
	// assembler code replaces default loop of 4 instructions with loop of 3 instructions
	asm volatile(
		"and	%1, %1	\n\t"
		"breq	.+6	\n\t"
		"add	%0, %0	\n\t"
		"dec	%1	\n\t"
		"brne	.-6	\n\t"
		: "=r" (mask), "=r" (bits)
		: "0"  (mask), "1"  (bits)
	);
	
	if((*ptr1 | mask) != (*ptr2 | mask)) return 1;
	
	return 0;
}

